/*
* File:     alias.c
* Author:   Michael Chinn
* UDel ID:  mechinn
*/

/*
 * Should work the same as the tcsh builtin one. When ran with no
 * arguments prints the aliases the shell knows about. When ran with
 * arguments it should install an alias into the alias table the shell
 * knows about. Additionally, the shell needs to be able to run your
 * aliases to receive full credit. Supply your own test cases to show
 * that it works
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "alias.h"

//main alias function
void alias(int argc, char **argv) {
    if (argc==1) { //if just alias print all aliases
        if(!aliasHead) {
            printf("No Aliases\n"); //if there is no aliases tell the user that
        } else {
            //go through the alias list and print
            struct aliasList *temp = aliasHead;
            while(temp){
                printf("%s == %s",temp->name,temp->command);
                temp = temp->next;
            }
        }
    } else if (argc==3) { // if "alias name command" make the name equal to the command in the alias table
        struct aliasList *temp;

        temp=(struct aliasList *)malloc(sizeof(struct aliasList));
        temp->name = strdup(argv[1]);
        temp->command = strdup(argv[2]);

        if (aliasHead == NULL){
            //List is Empty
            aliasHead=temp;
            aliasHead->next=NULL;
        } else {
            temp->next=aliasHead;
            aliasHead=temp;
        }

        printf("%s == %s",temp->name,temp->command);
    } else { //anything else they did it wrong
        printf("Usage: alias aliasName aliasCommand");
    }
}
//free alias memory
void freeAlias() {
   struct aliasList *temp;
   while( aliasHead != NULL ) {
      temp = aliasHead->next;
      free( aliasHead );
      aliasHead = temp;
   }
}
//replace user args that are aliases with their real meaning
void replaceAlias(int argc, char** argv) {
    struct aliasList *temp = aliasHead;
    int i;
    while(temp) { //if the current alias we are on exists
        for(i=0;i<argc;++i){//for each argument
            if(!strncmp(argv[i],temp->name,strlen(temp->name))){ //check if the argument is an alias
                argv[i] = temp->command; //if so replace the argument with its real value
            }
        }
        temp = temp->next; //move to the next alias in the list
    }
}
