/*
* File:   external.c
* Author: Michael Chinn
* UDel ID: mechinn
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <glob.h>

#include "external.h"
#include "get_path.h"

int externalBackground(int argc,char **argv, char** envp) {
    //setup varaibles
    char *command;
    int i;
    pid_t cpid = fork();				/* create a new process */
    struct pathelement *pathlist;
    pathlist = get_path();
    /* Both the old and new process will run the following code if successful. */
    if ( cpid == -1 ){
        perror("fork");
        exit(-1);
    } else if ( cpid == 0 ){ // in child process
        signal(SIGINT, SIG_DFL); //allow interrupts again temporarily
        //check for direct path
        if (!strncmp(argv[0],"/",1)||!strncmp(argv[0],"./",2)||!strncmp(argv[0],"../",3)) {
            if (access(argv[0], X_OK) == -1) { //check if we are allowed to execute it
                perror( "access" );
            } else {
                printf ("Executing %s\n",argv[0]);
                if ( execve(argv[0], argv, envp) == -1 ) { //execute it
                    perror( "execve" );
                }
            }
            freePath(pathlist); //free up path memory
            exit(0);
        } else {
            if (argc>1){
                // handle wildcards
                glob_t globbuf;
                globbuf.gl_offs = 1;
                glob(argv[1], GLOB_DOOFFS | GLOB_NOCHECK, NULL, &globbuf);//glob wildcard
                for(i=2;i<argc;++i){
                    glob(argv[i], GLOB_DOOFFS | GLOB_APPEND | GLOB_NOCHECK, NULL, &globbuf); //append wildcard
                }
                globbuf.gl_pathv[0] = argv[0]; //put the command at the front
                // execute line after fixing wildcards
                if ( execvp(argv[0], &globbuf.gl_pathv[0]) == -1 ) {
                    perror( "execve" );
                }
            } else {
                if ( execvp(argv[0], &argv[0]) == -1 ) {
                    perror( "execve" );
                }
            }
            freePath(pathlist); //free up path memory
            exit(0);
        }
        freePath(pathlist); //free up path memory
        exit(1);
    } else {
        signal(SIGINT, SIG_IGN); //dont allow interrupts just in case
        if (waitpid(cpid, NULL, WNOHANG) == -1) { //wait for child process to finish
            perror("waitpid");
            exit(1);
        }
    }
}

int external(int argc,char **argv, char** envp) {
    //setup varaibles
    char *command;
    int status,i;
    pid_t cpid = fork();				/* create a new process */
    struct pathelement *pathlist;
    pathlist = get_path();
    /* Both the old and new process will run the following code if successful. */
    if ( cpid == -1 ){
        perror("fork");
        exit(-1);
    } else if ( cpid == 0 ){ // in child process
        signal(SIGINT, SIG_DFL); //allow interrupts again temporarily
        //check for direct path
        if (!strncmp(argv[0],"/",1)||!strncmp(argv[0],"./",2)||!strncmp(argv[0],"../",3)) {
            if (access(argv[0], X_OK) == -1) { //check if we are allowed to execute it
                perror( "access" );
            } else {
                printf ("Executing %s\n",argv[0]);
                if ( execve(argv[0], argv, envp) == -1 ) { //execute it
                    perror( "execve" );
                }
            }
            freePath(pathlist); //free up path memory
            exit(0);
        } else {
            if (argc>1){
                // handle wildcards
                glob_t globbuf;
                globbuf.gl_offs = 1;
                glob(argv[1], GLOB_DOOFFS | GLOB_NOCHECK, NULL, &globbuf);//glob wildcard
                for(i=2;i<argc;++i){
                    glob(argv[i], GLOB_DOOFFS | GLOB_APPEND | GLOB_NOCHECK, NULL, &globbuf); //append wildcard
                }
                globbuf.gl_pathv[0] = argv[0]; //put the command at the front
                // execute line after fixing wildcards
                if ( execvp(argv[0], &globbuf.gl_pathv[0]) == -1 ) {
                    perror( "execve" );
                }
            } else {
                if ( execvp(argv[0], &argv[0]) == -1 ) {
                    perror( "execve" );
                }
            }
            freePath(pathlist); //free up path memory
            exit(0);
        }
        freePath(pathlist); //free up path memory
        exit(1);
    } else { //in parent process
        signal(SIGINT, SIG_IGN); //dont allow interrupts just in case
        if (waitpid(cpid, &status, WUNTRACED | WCONTINUED) == -1) { //wait for child process to finish
            perror("waitpid");
            exit(1);
        }
        if (WEXITSTATUS(status)!=0) {
            printf("could not execute command\n");
        }
    }
}

int externalPiped(int argc,char **argv, char** envp) {
    //setup varaibles
    int i;
    struct pathelement *pathlist;
    pathlist = get_path();
    signal(SIGINT, SIG_DFL); //allow interrupts again temporarily
    //check for direct path
    if (!strncmp(argv[0],"/",1)||!strncmp(argv[0],"./",2)||!strncmp(argv[0],"../",3)) {
        if (access(argv[0], X_OK) == -1) { //check if we are allowed to execute it
            perror( "access" );
        } else {
            printf ("Executing %s\n",argv[0]);
            if ( execve(argv[0], argv, envp) == -1 ) { //execute it
                perror( "execve" );
            }
        }
    } else {
        if (argc>1){
            // handle wildcards
            glob_t globbuf;
            globbuf.gl_offs = 1;
            glob(argv[1], GLOB_DOOFFS | GLOB_NOCHECK, NULL, &globbuf);//glob wildcard
            for(i=2;i<argc;++i){
                glob(argv[i], GLOB_DOOFFS | GLOB_APPEND | GLOB_NOCHECK, NULL, &globbuf); //append wildcard
            }
            globbuf.gl_pathv[0] = argv[0]; //put the command at the front
            // execute line after fixing wildcards
            if ( execvp(argv[0], &globbuf.gl_pathv[0]) == -1 ) {
                perror( "execve" );
            }
        } else {
            if ( execvp(argv[0], &argv[0]) == -1 ) {
                perror( "execve" );
            }
        }
    }
    freePath(pathlist); //free up path memory
    exit(0);
}
