/*
* File:     setenv.c
* Author:   Michael Chinn
* UDel ID:  mechinn
*/

/*
 * Should work the same as the tcsh builtin one. When ran with no
 * arguments prints the whole environment, the same as a printenv.
 * When ran with one argument sets that as an empty environment
 * variable. When ran with two arguments the second one is the value
 * of the first. When ran with more args print the same error message
 * to stderr that tcsh does. You can use the putenv(3C) function for
 * this command. Special care must be given when PATH and HOME are
 * changed. When PATH is changed be sure to update your linked list
 * for the path directories (and free() up the old one). When HOME is
 * changed cd with no arguments should now go to the new home. Provide
 * your own test runs to show this works
 */

#include <stdio.h>
#include <stdlib.h>

#include "setenv.h"
#include "printenv.h"
// main setenv
void setEnv(int argc, char **argv, char** envp) {
    if (argc==1) { //if just setenv then print the entire environment
        printEnv(argc,argv,envp);
    } else if (argc==2) { // create env variable
        if(putenv(argv[1])) {
            perror("putenv");
        }
    } else if (argc==3) { // create env variable and make variable equal to second arg
        if(setenv(argv[1], argv[2], 1) == -1) {
            perror("setenv");
        }
    } else {
        printf("Usage: setenv <environmentVariable> <variableValue>");
    }
}
