/*
* File:     sh.c
* Author:   Michael Chinn
* UDel ID:  mechinn
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <signal.h>
#include <sys/wait.h>
#include <readline.h>
#include <history.h>

#include "get_path.h"
#include "which.h"
#include "where.h"
#include "list.h"
#include "process.h"
#include "printenv.h"
#include "alias.h"
#include "myHistory.h"
#include "setenv.h"
#include "external.h"
#include "help.h"
#include "procstat.h"
#include "watchmail.h"

#define length 512 // max input length allowed

/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

/* Read a string, and return a pointer to it.
 * Returns NULL on EOF. */
char *rl_gets (char* prompts) {
      /* If the buffer has already been allocated,
       * return the memory to the free pool. */
    if (line_read) {
      free (line_read);
      line_read = (char *)NULL;
    }

    /* Get a line from the user. */
    line_read = readline (prompts);

    /* If the line has any text in it,
    * save it on the history. */
    if (line_read && *line_read){
        add_history (line_read);
    }

    return (line_read);
}

int main( int argc, char **argv, char** envp) {
    // initialize variables
    pid_t checkzomb;
    int fd[2];
    pid_t childpid;
    int fid,out,in;
    int noclobber = 0;
    int temp;
    int redirect,piped;
    int i = -1;
    char input [length];
    char *parsed[(length/2)+1];
    int newArgC = -1;
    char defprompt[] = " [";
    char defpromptend[] = "]> ";
    char* prompts = malloc(length*sizeof(char));
    int path_max = pathconf("/", _PC_PATH_MAX) + 1;

    // get current working directory
    char cwd[path_max]; //current working directory
    char lastcwd[path_max]; //last working directory
    char tmp[path_max]; //temp array to store working directly in when going back to the last working directory
    if (!getcwd(cwd, path_max)) {
        perror("failed to get current working directory");
        exit(1);
    }
    strcpy(prompts,defprompt);
    strcat(prompts,cwd);
    strcat(prompts,defpromptend);

    // main loop, never should exit unless exit is called by user
    while(1) {
        // reset the stdin, stdout, stderr file descriptors each time through loop
        piped = 0;
        redirect = 0;
        out = open("/dev/tty",O_WRONLY);
        in = open("/dev/tty",O_RDONLY);
        close(STDIN_FILENO);
        dup(in);
        close(STDOUT_FILENO);
        dup(out);
        close(STDERR_FILENO);
        dup(out);
        close(out);
        close(in);
        // ignore signals
        signal(SIGINT, SIG_IGN);
        signal(SIGTSTP, SIG_IGN);
        signal(SIGTERM, SIG_IGN);
        // command prompt
        //printf (prompts, cwd);
        //fgets(input,length,stdin);

        strncpy(input,rl_gets(prompts),length);
        
        addMyHistory(input); //add input to history

        // parse input
        newArgC = 0;
        parsed[newArgC] = strtok(input, " \n");
        while( parsed[newArgC] != NULL ) {
            ++newArgC;
            parsed[newArgC] = strtok(NULL," \n");
        }

        if(newArgC>0){
            // create argv based off argc
            char* newArgV[newArgC+1];
            for (i=0;i<=newArgC;++i){
                newArgV[i] = parsed[i];
            }
            if (newArgC>2){
                if (!strncmp(newArgV[newArgC-2],">>&",3)){
                    //appending redirect stdout & stderr
                    redirect=1;
                    if(noclobber){
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_APPEND, S_IRWXU);
                        if(fid==-1){
                            perror("open");
                        } else {
                            close(STDOUT_FILENO);
                            close(STDERR_FILENO);
                            dup(fid);
                            dup(fid);
                            close(fid);
                            newArgC-=2;
                            newArgV[newArgC]=NULL;
                        }
                    } else {
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_APPEND, S_IRWXU);
                        close(STDOUT_FILENO);
                        close(STDERR_FILENO);
                        dup(fid);
                        dup(fid);
                        close(fid);
                        newArgC-=2;
                        newArgV[newArgC]=NULL;
                    }
                } else if (!strncmp(newArgV[newArgC-2],">>",2)){
                    //appending redirect stdout
                    redirect=1;
                    if(noclobber){
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_APPEND, S_IRWXU);
                        if(fid==-1){
                            perror("open");
                        } else {
                            close(STDOUT_FILENO);
                            dup(fid);
                            close(fid);
                            newArgC-=2;
                            newArgV[newArgC]=NULL;
                        }
                    } else {
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_APPEND, S_IRWXU);
                        close(STDOUT_FILENO);
                        dup(fid);
                        close(fid);
                        newArgC-=2;
                        newArgV[newArgC]=NULL;
                    }
                } else if (!strncmp(newArgV[newArgC-2],">&",2)){
                    //redirect stdout & stderr
                    redirect=1;
                    if(noclobber){
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_EXCL|O_TRUNC, S_IRWXU);
                        if(fid==-1) {

                        } else {
                            close(STDOUT_FILENO);
                            close(STDERR_FILENO);
                            dup(fid);
                            dup(fid);
                            close(fid);
                            newArgC-=2;
                            newArgV[newArgC]=NULL;
                        }
                    } else {
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU);
                        close(STDOUT_FILENO);
                        close(STDERR_FILENO);
                        dup(fid);
                        dup(fid);
                        close(fid);
                        newArgC-=2;
                        newArgV[newArgC]=NULL;
                    }
                } else if (!strncmp(newArgV[newArgC-2],">",1)){
                    //redirect stdout
                    redirect=1;
                    if(noclobber){
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_EXCL|O_TRUNC, S_IRWXU);
                        if(fid==-1){
                            perror("open");
                        } else {
                            close(STDOUT_FILENO);
                            dup(fid);
                            close(fid);
                            newArgC-=2;
                            newArgV[newArgC]=NULL;
                        }
                    } else {
                        fid = open(newArgV[newArgC-1],O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU);
                        close(STDOUT_FILENO);
                        dup(fid);
                        close(fid);
                        newArgC-=2;
                        newArgV[newArgC]=NULL;
                    }
                } else if (!strncmp(newArgV[newArgC-2],"<",1)){
                    //redirect stdin to file
                    fid = open(newArgV[newArgC-1],O_RDONLY);
                    close(STDIN_FILENO);
                    dup(fid);
                    close(fid);
                    newArgC-=2;
                    newArgV[newArgC]=NULL;
                } else {
                    //check for pipe
                    for(i=1;i<newArgC-1;++i){
                        if(!strncmp(newArgV[i],"|&",2)){
                            //pipe stdout & stderr
                            pipe(fd);
                            if((childpid = fork()) == -1) {
                                perror("fork");
                                break; //get out of the for loop
                            }
                            if(childpid==0) {
                                close(fd[0]);
                                close(STDOUT_FILENO);
                                close(STDERR_FILENO);
                                dup(fd[1]);
                                dup(fd[1]);
                                temp = i-newArgC;
                                newArgC+=temp;
                                newArgV[newArgC]=NULL;
                                piped = 1;
                            } else {
                                close(fd[1]);
                                close(STDIN_FILENO);
                                dup(fd[0]);
                                temp = 0;
                                for(++i;i<newArgC;++i){
                                    newArgV[temp] = newArgV[i];
                                    ++temp;
                                }
                                newArgV[temp]=NULL;
                                newArgC=temp;
                                piped = 2;
                                wait(NULL);
                            }
                            break;
                        } else if(!strncmp(newArgV[i],"|",1)){
                            //pipe stdout
                            pipe(fd);
                            if((childpid = fork()) == -1) {
                                perror("fork");
                                break; //get out of the for loop
                            }
                            if(childpid==0) {
                                close(fd[0]);
                                close(STDOUT_FILENO);
                                dup(fd[1]);
                                temp = i-newArgC;
                                newArgC+=temp;
                                newArgV[newArgC]=NULL;
                                piped = 1;
                            } else {
                                close(fd[1]);
                                close(STDIN_FILENO);
                                dup(fd[0]);
                                temp = 0;
                                for(++i;i<newArgC;++i){
                                    newArgV[temp] = newArgV[i];
                                    ++temp;
                                }
                                newArgV[temp]=NULL;
                                newArgC=temp;
                                piped = 2;
                                wait(NULL);
                            }
                            break; //get out of for loop, we found a pipe dont check anymore
                        }
                    }
                }
            }

            // built in commands
            if (!strncmp(newArgV[0],"help",4)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin help\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                help(newArgC,newArgV);
            } else if (!strncmp(newArgV[0],"exit",4)) {
                if (piped!=1&&redirect!=1){
                printf ("Executing builtin exit\n");
                }
                freeMyHistory();  //free memory from history linked list
                freeAlias();    //free aliases from alias linked list
                freeMail();     //free watchmails from mail linked list
                exit(0);        //good exit
            } else if (!strncmp(newArgV[0],"which",5)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin which\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                which(newArgC,newArgV, envp);
            } else if (!strncmp(newArgV[0],"where",5)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin where\n");
                }
                where(newArgC,newArgV,envp);
            } else if (!strncmp(newArgV[0],"cd",2)) {
                /*
                 * cd - chdir(2) to directory given, with no arguments chdir to home dir, with a '-'
                 * as the only argument chdirs to directory previously in, the same as tcsh does
                 */
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin cd\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                // change directory
                if (newArgC==1) { //if just "cd"
                    if(chdir(getenv("HOME"))==-1) { // change directory
                        perror( "cd" );
                    } else {
                        strcpy(lastcwd,cwd);//copy the current directory into the last directory holder
                    }
                    if (!getcwd(cwd, path_max)) { // get current path
                        perror("failed to get current working directory");
                        exit(1);
                    } else {
                        strcpy(prompts,defprompt);
                        strcat(prompts,cwd);
                        strcat(prompts,defpromptend);
                    }
                } else if (newArgC==2) { //if "cd path"
                    if(strlen(newArgV[1])==1&&!strncmp(newArgV[1],"-",1)) { //if "cd -" return to last directory
                        strcpy(tmp,cwd);
                        if(chdir(lastcwd)==-1) {
                            printf("no previous directory");
                        } else {
                            strcpy(lastcwd,tmp);
                        }
                    } else {
                        if(chdir(newArgV[1])==-1) {
                            perror( "cd" );
                        } else {
                            strcpy(lastcwd,cwd);
                        }
                    }
                    if (!getcwd(cwd, path_max)) {
                        perror("failed to get current working directory");
                        exit(1);
                    } else {
                        strcpy(prompts,defprompt);
                        strcat(prompts,cwd);
                        strcat(prompts,defpromptend);
                    }
                } else { //user has too many args
                    printf("Usage: cd <path/->");
                }
            } else if (!strncmp(newArgV[0],"pwd",3)) {
                // print the current working dir.
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin pwd\n");
                }
                printf ("Current Working Directory:\n%s\n", cwd); //just print the cwd
            } else if (!strncmp(newArgV[0],"list",4)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin list\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                list(newArgC,newArgV);
            } else if (!strncmp(newArgV[0],"pid",3)) {
                // prints the pid of the shell
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin pid\n");
                }
                printf("PID: %d\n",getpid()); //just get print the pid
            } else if (!strncmp(newArgV[0],"kill",4)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin kill\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                process(newArgC,newArgV);
            } else if (!strncmp(newArgV[0],"prompt",6)) {
                /*
                 * When ran with no arguments prompts for a new prompt prefix string.
                 * When given an argument make that the new prompt prefix
                 */
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin prompt\n");
                }
                if(newArgC==1) { //if just prompt
                    printf("Prompt Prefix: \n"); //ask the user for what the prompt prefix should be
                    fgets(input,length,stdin);
                    strcpy(prompts,strtok(input,"\n"));//copy the new prompt prfix into prompt string
                } if(newArgC==2) { //if "prompt prefix"
                    strcpy(prompts,strdup(newArgV[1]));// just do the copy from arg
                } else {
                    printf("Usage: prompt <promptPrefix>");
                }
            } else if (!strncmp(newArgV[0],"printenv",8)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin printenv\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                printEnv(newArgC,newArgV,envp);
            } else if (!strncmp(newArgV[0],"alias",5)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin alias\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                alias(newArgC,newArgV);
            } else if (!strncmp(newArgV[0],"history",7)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin history\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                myHistory(newArgC,newArgV);
            } else if (!strncmp(newArgV[0],"setenv",6)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin setenv\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                setEnv(newArgC,newArgV,envp);
            } else if (!strncmp(newArgV[0],"procstat",8)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin procstat\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                procstat(newArgC,newArgV,envp);
            } else if (!strncmp(newArgV[0],"watchmail",9)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin watchmail\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                watchmail(newArgC,newArgV,envp);
            } else if (!strncmp(newArgV[0],"noclobber",9)) {
                if (piped!=1&&redirect!=1){
                    printf ("Executing builtin noclobber\n");
                }
                if(noclobber){
                    noclobber = 0;
                } else {
                    noclobber = 1;
                }
                printf ("noclobber changed to: %d\n",noclobber);
            } else {
                if (piped!=1&&redirect!=1){
                    printf ("Executing external command\n");
                }
                replaceAlias(newArgC,newArgV); //replace inputed aliases with their real meaning
                if (piped==1) {
                    externalPiped(newArgC, newArgV, envp);
                }else if (!strncmp(newArgV[newArgC-1],"&",1)) {
                    newArgV[newArgC-1]=NULL;
                    externalBackground(newArgC-1, newArgV, envp);
                } else {
                    external(newArgC, newArgV, envp);
                }
            }
        }
        while ((checkzomb = waitpid(-1, NULL, WNOHANG)) > 0 );
        if(piped==1) {
            close(fd[1]);
            exit(0);
        } else if(piped==2){
            close(fd[0]);
        }
    }
} /* end main() */
